# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import contextlib
from itertools import chain
import re

import requests

from ..attachment.base import AttachBase
from ..common import NotifyFormat, NotifyType
from ..conversion import convert_between
from ..locale import gettext_lazy as _
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase

# Flag used as a placeholder to sending to all devices
PUSHOVER_SEND_TO_ALL = "ALL_DEVICES"

# Used to detect a Device
VALIDATE_DEVICE = re.compile(r"^\s*(?P<device>[a-z0-9_-]{1,25})\s*$", re.I)


# Priorities
class PushoverPriority:
    LOW = -2
    MODERATE = -1
    NORMAL = 0
    HIGH = 1
    EMERGENCY = 2


# Sounds
class PushoverSound:
    PUSHOVER = "pushover"
    BIKE = "bike"
    BUGLE = "bugle"
    CASHREGISTER = "cashregister"
    CLASSICAL = "classical"
    COSMIC = "cosmic"
    FALLING = "falling"
    GAMELAN = "gamelan"
    INCOMING = "incoming"
    INTERMISSION = "intermission"
    MAGIC = "magic"
    MECHANICAL = "mechanical"
    PIANOBAR = "pianobar"
    SIREN = "siren"
    SPACEALARM = "spacealarm"
    TUGBOAT = "tugboat"
    ALIEN = "alien"
    CLIMB = "climb"
    PERSISTENT = "persistent"
    ECHO = "echo"
    UPDOWN = "updown"
    NONE = "none"


PUSHOVER_SOUNDS = (
    PushoverSound.PUSHOVER,
    PushoverSound.BIKE,
    PushoverSound.BUGLE,
    PushoverSound.CASHREGISTER,
    PushoverSound.CLASSICAL,
    PushoverSound.COSMIC,
    PushoverSound.FALLING,
    PushoverSound.GAMELAN,
    PushoverSound.INCOMING,
    PushoverSound.INTERMISSION,
    PushoverSound.MAGIC,
    PushoverSound.MECHANICAL,
    PushoverSound.PIANOBAR,
    PushoverSound.SIREN,
    PushoverSound.SPACEALARM,
    PushoverSound.TUGBOAT,
    PushoverSound.ALIEN,
    PushoverSound.CLIMB,
    PushoverSound.PERSISTENT,
    PushoverSound.ECHO,
    PushoverSound.UPDOWN,
    PushoverSound.NONE,
)

PUSHOVER_PRIORITIES = {
    # Note: This also acts as a reverse lookup mapping
    PushoverPriority.LOW: "low",
    PushoverPriority.MODERATE: "moderate",
    PushoverPriority.NORMAL: "normal",
    PushoverPriority.HIGH: "high",
    PushoverPriority.EMERGENCY: "emergency",
}

PUSHOVER_PRIORITY_MAP = {
    # Maps against string 'low'
    "l": PushoverPriority.LOW,
    # Maps against string 'moderate'
    "m": PushoverPriority.MODERATE,
    # Maps against string 'normal'
    "n": PushoverPriority.NORMAL,
    # Maps against string 'high'
    "h": PushoverPriority.HIGH,
    # Maps against string 'emergency'
    "e": PushoverPriority.EMERGENCY,
    # Entries to additionally support (so more like Pushover's API)
    "-2": PushoverPriority.LOW,
    "-1": PushoverPriority.MODERATE,
    "0": PushoverPriority.NORMAL,
    "1": PushoverPriority.HIGH,
    "2": PushoverPriority.EMERGENCY,
}

# Extend HTTP Error Messages
PUSHOVER_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Token.",
}


class NotifyPushover(NotifyBase):
    """A wrapper for Pushover Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Pushover"

    # The services URL
    service_url = "https://pushover.net/"

    # All pushover requests are secure
    secure_protocol = "pover"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_pushover"

    # Pushover uses the http protocol with JSON requests
    notify_url = "https://api.pushover.net/1/messages.json"

    # Support attachments
    attachment_support = True

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 1024

    # Default Pushover sound
    default_pushover_sound = PushoverSound.PUSHOVER

    # 5MB is the maximum supported image filesize as per documentation
    # here: https://pushover.net/api#limits (Oct 5th, 2025)
    attach_max_size_bytes = 5242880

    # The regular expression of the current attachment supported mime types
    # At this time it is only images
    attach_supported_mime_type = r"^image/.*"

    # Define object templates
    templates = (
        "{schema}://{user_key}@{token}",
        "{schema}://{user_key}@{token}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user_key": {
                "name": _("User Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "token": {
                "name": _("Access Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_device": {
                "name": _("Target Device"),
                "type": "string",
                "regex": (r"^[a-z0-9_-]{1,25}$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "priority": {
                "name": _("Priority"),
                "type": "choice:int",
                "values": PUSHOVER_PRIORITIES,
                "default": PushoverPriority.NORMAL,
            },
            "sound": {
                "name": _("Sound"),
                "type": "string",
                "regex": (r"^[a-z]{1,12}$", "i"),
                "default": PushoverSound.PUSHOVER,
            },
            "url": {
                "name": _("URL"),
                "map_to": "supplemental_url",
                "type": "string",
            },
            "url_title": {
                "name": _("URL Title"),
                "map_to": "supplemental_url_title",
                "type": "string",
            },
            "retry": {
                "name": _("Retry"),
                "type": "int",
                "min": 30,
                "default": 900,  # 15 minutes
            },
            "expire": {
                "name": _("Expire"),
                "type": "int",
                "min": 0,
                "max": 10800,
                "default": 3600,  # 1 hour
            },
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(
        self,
        user_key,
        token,
        targets=None,
        priority=None,
        sound=None,
        retry=None,
        expire=None,
        supplemental_url=None,
        supplemental_url_title=None,
        **kwargs,
    ):
        """Initialize Pushover Object."""
        super().__init__(**kwargs)

        # Access Token (associated with project)
        self.token = validate_regex(token)
        if not self.token:
            msg = f"An invalid Pushover Access Token ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # User Key (associated with project)
        self.user_key = validate_regex(user_key)
        if not self.user_key:
            msg = f"An invalid Pushover User Key ({user_key}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Track our valid devices
        targets = parse_list(targets)

        # Track any invalid entries
        self.invalid_targets = []

        if len(targets) == 0:
            self.targets = (PUSHOVER_SEND_TO_ALL,)

        else:
            self.targets = []
            for target in targets:
                result = VALIDATE_DEVICE.match(target)
                if result:
                    # Store device information
                    self.targets.append(result.group("device"))
                    continue

                self.logger.warning(
                    f"Dropped invalid Pushover device ({target}) specified.",
                )
                self.invalid_targets.append(target)

        # Setup supplemental url
        self.supplemental_url = supplemental_url
        self.supplemental_url_title = supplemental_url_title

        # Setup our sound
        self.sound = (
            NotifyPushover.default_pushover_sound
            if not isinstance(sound, str)
            else sound.lower()
        )
        if self.sound and self.sound not in PUSHOVER_SOUNDS:
            msg = f"Using custom sound specified ({sound}). "
            self.logger.debug(msg)

        # The Priority of the message
        self.priority = int(
            NotifyPushover.template_args["priority"]["default"]
            if priority is None
            else next(
                (
                    v
                    for k, v in PUSHOVER_PRIORITY_MAP.items()
                    if str(priority).lower().startswith(k)
                ),
                NotifyPushover.template_args["priority"]["default"],
            )
        )

        # The following are for emergency alerts
        if self.priority == PushoverPriority.EMERGENCY:

            # How often to resend notification, in seconds
            self.retry = self.template_args["retry"]["default"]
            with contextlib.suppress(ValueError, TypeError):
                # Get our retry value
                self.retry = int(retry)

            # How often to resend notification, in seconds
            self.expire = self.template_args["expire"]["default"]
            with contextlib.suppress(ValueError, TypeError):
                # Acquire our expiry value
                self.expire = int(expire)

            if self.retry < 30:
                msg = "Pushover retry must be at least 30 seconds."
                self.logger.warning(msg)
                raise TypeError(msg)

            if self.expire < 0 or self.expire > 10800:
                msg = (
                    "Pushover expire must reside in the range of "
                    "0 to 10800 seconds."
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Pushover Notification."""

        if not self.targets:
            # There were no services to notify
            self.logger.warning("There were no Pushover targets to notify.")
            return False

        # prepare JSON Object
        payload = {
            "token": self.token,
            "user": self.user_key,
            "priority": str(self.priority),
            "title": title if title else self.app_desc,
            "message": body,
            "device": ",".join(self.targets),
            "sound": self.sound,
        }

        if self.supplemental_url:
            payload["url"] = self.supplemental_url

        if self.supplemental_url_title:
            payload["url_title"] = self.supplemental_url_title

        if self.notify_format == NotifyFormat.HTML:
            # https://pushover.net/api#html
            payload["html"] = 1

        elif self.notify_format == NotifyFormat.MARKDOWN:
            payload["message"] = convert_between(
                NotifyFormat.MARKDOWN, NotifyFormat.HTML, body
            )
            payload["html"] = 1

        if self.priority == PushoverPriority.EMERGENCY:
            payload.update({"retry": self.retry, "expire": self.expire})

        if attach and self.attachment_support:
            # Create a copy of our payload
            _payload = payload.copy()

            # Send with attachments
            for no, attachment in enumerate(attach):
                if no or not body:
                    # To handle multiple attachments, clean up our message
                    _payload["message"] = attachment.name

                if not self._send(_payload, attachment):
                    # Mark our failure
                    return False

                # Clear our title if previously set
                _payload["title"] = ""

                # No need to alarm for each consecutive attachment uploaded
                # afterwards
                _payload["sound"] = PushoverSound.NONE

        else:
            # Simple send
            return self._send(payload)

        return True

    def _send(self, payload, attach=None):
        """Wrapper to the requests (post) object."""

        if isinstance(attach, AttachBase):
            # Perform some simple error checking
            if not attach:
                # We could not access the attachment
                self.logger.error(
                    f"Could not access attachment {attach.url(privacy=True)}."
                )
                return False

            # Perform some basic checks as we want to gracefully skip
            # over unsupported mime types.
            if not re.match(
                self.attach_supported_mime_type, attach.mimetype, re.I
            ):
                # No problem; we just don't support this attachment
                # type; gracefully move along
                self.logger.debug(
                    "Ignored unsupported Pushover attachment"
                    f" ({attach.mimetype}): {attach.url(privacy=True)}"
                )

                attach = None

            else:
                # If we get here, we're dealing with a supported image.
                # Verify that the filesize is okay though.
                file_size = len(attach)
                if not (
                    file_size > 0 and file_size <= self.attach_max_size_bytes
                ):

                    # File size is no good
                    self.logger.warning(
                        f"Pushover attachment size ({file_size}B) exceeds"
                        f" limit: {attach.url(privacy=True)}"
                    )

                    return False

                self.logger.debug(
                    f"Posting Pushover attachment {attach.url(privacy=True)}"
                )

        # Default Header
        headers = {
            "User-Agent": self.app_id,
        }

        # Authentication
        auth = (self.token, "")

        # Some default values for our request object to which we'll update
        # depending on what our payload is
        files = None

        self.logger.debug(
            "Pushover POST URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Pushover Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            # Open our attachment path if required:
            if attach:
                files = {
                    "attachment": (
                        attach.name,
                        # file handle is safely closed in `finally`; inline
                        # open is intentional
                        open(attach.path, "rb"),  # noqa: SIM115
                    )}

            r = requests.post(
                self.notify_url,
                data=payload,
                headers=headers,
                files=files,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyPushover.http_response_code_lookup(
                    r.status_code, PUSHOVER_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send Pushover notification to {}: "
                    "{}{}error={}.".format(
                        payload["device"],
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                return False

            else:
                self.logger.info(
                    "Sent Pushover notification to {}.".format(
                        payload["device"]
                    )
                )

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Pushover:{} ".format(
                    payload["device"]
                )
                + "notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            return False

        except OSError as e:
            self.logger.warning(
                "An I/O error occurred while reading {}.".format(
                    attach.name if attach else "attachment"
                )
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            return False

        finally:
            # Close our file (if it's open) stored in the second element
            # of our files tuple (index 1)
            if files:
                files["attachment"][1].close()

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.user_key, self.token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "priority": (
                PUSHOVER_PRIORITIES[self.template_args["priority"]["default"]]
                if self.priority not in PUSHOVER_PRIORITIES
                else PUSHOVER_PRIORITIES[self.priority]
            ),
        }

        # Only add expire and retry for emergency messages,
        # pushover ignores for all other priorities
        if self.priority == PushoverPriority.EMERGENCY:
            params.update({"expire": self.expire, "retry": self.retry})

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Escape our devices
        devices = "/".join([
            NotifyPushover.quote(x, safe="")
            for x in chain(self.targets, self.invalid_targets)
        ])

        if devices == PUSHOVER_SEND_TO_ALL:
            # keyword is reserved for internal usage only; it's safe to remove
            # it from the devices list
            devices = ""

        return "{schema}://{user_key}@{token}/{devices}/?{params}".format(
            schema=self.secure_protocol,
            user_key=self.pprint(self.user_key, privacy, safe=""),
            token=self.pprint(self.token, privacy, safe=""),
            devices=devices,
            params=NotifyPushover.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Set our priority
        if "priority" in results["qsd"] and len(results["qsd"]["priority"]):
            results["priority"] = NotifyPushover.unquote(
                results["qsd"]["priority"]
            )

        # Retrieve all of our targets
        results["targets"] = NotifyPushover.split_path(results["fullpath"])

        # User Key is retrieved from the user
        results["user_key"] = NotifyPushover.unquote(results["user"])

        # Get the sound
        if "sound" in results["qsd"] and len(results["qsd"]["sound"]):
            results["sound"] = NotifyPushover.unquote(results["qsd"]["sound"])

        # Get the supplementary url
        if "url" in results["qsd"] and len(results["qsd"]["url"]):
            results["supplemental_url"] = NotifyPushover.unquote(
                results["qsd"]["url"]
            )
        if "url_title" in results["qsd"] and len(results["qsd"]["url_title"]):
            results["supplemental_url_title"] = results["qsd"]["url_title"]

        # Get expire and retry
        if "expire" in results["qsd"] and len(results["qsd"]["expire"]):
            results["expire"] = results["qsd"]["expire"]
        if "retry" in results["qsd"] and len(results["qsd"]["retry"]):
            results["retry"] = results["qsd"]["retry"]

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyPushover.parse_list(
                results["qsd"]["to"]
            )

        # Token
        results["token"] = NotifyPushover.unquote(results["host"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Gotify Docker configuration: https://hub.docker.com/r/gotify/server
# Example: https://github.com/gotify/server/blob/\
#      f2c2688f0b5e6a816bbcec768ca1c0de5af76b88/ADD_MESSAGE_EXAMPLES.md#python
# API: https://gotify.net/docs/swagger-docs

from json import dumps

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase


# Priorities
class GotifyPriority:
    LOW = 0
    MODERATE = 3
    NORMAL = 5
    HIGH = 8
    EMERGENCY = 10


GOTIFY_PRIORITIES = {
    # Note: This also acts as a reverse lookup mapping
    GotifyPriority.LOW: "low",
    GotifyPriority.MODERATE: "moderate",
    GotifyPriority.NORMAL: "normal",
    GotifyPriority.HIGH: "high",
    GotifyPriority.EMERGENCY: "emergency",
}

GOTIFY_PRIORITY_MAP = {
    # Maps against string 'low'
    "l": GotifyPriority.LOW,
    # Maps against string 'moderate'
    "m": GotifyPriority.MODERATE,
    # Maps against string 'normal'
    "n": GotifyPriority.NORMAL,
    # Maps against string 'high'
    "h": GotifyPriority.HIGH,
    # Maps against string 'emergency'
    "e": GotifyPriority.EMERGENCY,
    # Entries to additionally support (so more like Gotify's API)
    "10": GotifyPriority.EMERGENCY,
    # ^ 10 needs to be checked before '1' below or it will match the wrong
    # priority
    "0": GotifyPriority.LOW,
    "1": GotifyPriority.LOW,
    "2": GotifyPriority.LOW,
    "3": GotifyPriority.MODERATE,
    "4": GotifyPriority.MODERATE,
    "5": GotifyPriority.NORMAL,
    "6": GotifyPriority.NORMAL,
    "7": GotifyPriority.NORMAL,
    "8": GotifyPriority.HIGH,
    "9": GotifyPriority.HIGH,
}


class NotifyGotify(NotifyBase):
    """A wrapper for Gotify Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Gotify"

    # The services URL
    service_url = "https://github.com/gotify/server"

    # The default protocol
    protocol = "gotify"

    # The default secure protocol
    secure_protocol = "gotifys"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_gotify"

    # Disable throttle rate
    request_rate_per_sec = 0

    # Define object templates
    templates = (
        "{schema}://{host}/{token}",
        "{schema}://{host}:{port}/{token}",
        "{schema}://{host}{path}{token}",
        "{schema}://{host}:{port}{path}{token}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "path": {
                "name": _("Path"),
                "type": "string",
                "map_to": "fullpath",
                "default": "/",
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "priority": {
                "name": _("Priority"),
                "type": "choice:int",
                "values": GOTIFY_PRIORITIES,
                "default": GotifyPriority.NORMAL,
            },
        },
    )

    def __init__(self, token, priority=None, **kwargs):
        """Initialize Gotify Object."""
        super().__init__(**kwargs)

        # Token (associated with project)
        self.token = validate_regex(token)
        if not self.token:
            msg = f"An invalid Gotify Token ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # prepare our fullpath
        self.fullpath = kwargs.get("fullpath", "/")

        # The Priority of the message
        self.priority = int(
            NotifyGotify.template_args["priority"]["default"]
            if priority is None
            else next(
                (
                    v
                    for k, v in GOTIFY_PRIORITY_MAP.items()
                    if str(priority).lower().startswith(k)
                ),
                NotifyGotify.template_args["priority"]["default"],
            )
        )

        if self.secure:
            self.schema = "https"

        else:
            self.schema = "http"

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Gotify Notification."""

        url = f"{self.schema}://{self.host}"
        if self.port:
            url += f":{self.port}"

        # Append our remaining path
        url += f"{self.fullpath}message"

        # Prepare Gotify Object
        payload = {
            "priority": self.priority,
            "title": title,
            "message": body,
        }

        if self.notify_format == NotifyFormat.MARKDOWN:
            payload["extras"] = {
                "client::display": {"contentType": "text/markdown"}
            }

        # Our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "X-Gotify-Key": self.token,
        }

        self.logger.debug(
            f"Gotify POST URL: {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Gotify Payload: {payload!s}")

        # Always call throttle before the requests are made
        self.throttle()

        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyGotify.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Gotify notification: {}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Mark our failure
                return False

            else:
                self.logger.info("Sent Gotify notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Gotify "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Mark our failure
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port if self.port else (443 if self.secure else 80),
            self.fullpath.rstrip("/"),
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "priority": (
                GOTIFY_PRIORITIES[self.template_args["priority"]["default"]]
                if self.priority not in GOTIFY_PRIORITIES
                else GOTIFY_PRIORITIES[self.priority]
            ),
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Our default port
        default_port = 443 if self.secure else 80
        return "{schema}://{hostname}{port}{fullpath}{token}/?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            fullpath=NotifyGotify.quote(self.fullpath, safe="/"),
            token=self.pprint(self.token, privacy, safe=""),
            params=NotifyGotify.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early
            return results

        # Retrieve our escaped entries found on the fullpath
        entries = NotifyBase.split_path(results["fullpath"])

        # optionally find the provider key
        try:
            # The last entry is our token
            results["token"] = entries.pop()

        except IndexError:
            # No token was set
            results["token"] = None

        # Re-assemble our full path
        results["fullpath"] = (
            "/" if not entries else "/{}/".format("/".join(entries))
        )

        # Set our priority
        if "priority" in results["qsd"] and len(results["qsd"]["priority"]):
            results["priority"] = NotifyGotify.unquote(
                results["qsd"]["priority"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import contextlib
from datetime import timezone as _tz
from typing import Optional
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from ..logger import logger


def zoneinfo(name: str) -> Optional[ZoneInfo]:
    """
    More forgiving ZoneInfo instantiation
    - Accepts lower/upper case
    - Normalises common UTC variants
    """
    if not isinstance(name, str):
        return None

    raw = name.strip()
    if not raw:
        return None

    # Windows-safe: accept UTC family even without tzdata
    if raw.lower() in {
            "utc", "z", "gmt", "etc/utc", "etc/gmt", "gmt0", "utc0"}:
        return _tz.utc

    # Try exact match first
    try:
        return ZoneInfo(name)

    except ZoneInfoNotFoundError:
        pass

    # Try case-insensitive match across available keys
    from zoneinfo import available_timezones
    lowered = name.lower().strip()
    for zone in available_timezones():
        full_zone = zone.lower()
        if full_zone == lowered:
            return ZoneInfo(zone)

        with contextlib.suppress(IndexError):

            # Break our zones and enforce limit
            zones = full_zone.split("/")[1:3]

            # Possible we'll throw an index error here and that's okay
            location = zones[-1] if len(zones) == 1 else "/".join(zones)
            if location and location == lowered:
                return ZoneInfo(zone)

    logger.warning("Unknown timezone specified: %s", name)
    return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# For this to work correctly you need to create a webhook. To do this just
# click on the little gear icon next to the channel you're part of. From
# here you'll be able to access the Webhooks menu and create a new one.
#
#  When you've completed, you'll get a URL that looks a little like this:
#  https://discord.com/api/webhooks/417429632418316298/\
#         JHZ7lQml277CDHmQKMHI8qBe7bk2ZwO5UKjCiOAF7711o33MyqU344Qpgv7YTpadV_js
#
#  Simplified, it looks like this:
#     https://discord.com/api/webhooks/WEBHOOK_ID/WEBHOOK_TOKEN
#
#  This plugin will simply work using the url of:
#     discord://WEBHOOK_ID/WEBHOOK_TOKEN
#
# API Documentation on Webhooks:
#    - https://discord.com/developers/docs/resources/webhook
#
from datetime import datetime, timedelta, timezone
from itertools import chain
from json import dumps
import re

import requests

from ..attachment.base import AttachBase
from ..common import NotifyFormat, NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool, parse_list, validate_regex
from .base import NotifyBase

# Used to detect user/role IDs
USER_ROLE_DETECTION_RE = re.compile(
    r"\s*(?:<?@(?P<role>&?)(?P<id>[0-9]+)>?|@(?P<value>[a-z0-9]+))", re.I
)


class NotifyDiscord(NotifyBase):
    """A wrapper to Discord Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Discord"

    # The services URL
    service_url = "https://discord.com/"

    # The default secure protocol
    secure_protocol = "discord"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_discord"

    # Discord Webhook
    notify_url = "https://discord.com/api/webhooks"

    # Support attachments
    attachment_support = True

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_256

    # Discord is kind enough to return how many more requests we're allowed to
    # continue to make within it's header response as:
    # X-RateLimit-Reset: The epoc time (in seconds) we can expect our
    #                    rate-limit to be reset.
    # X-RateLimit-Remaining: an integer identifying how many requests we're
    #                        still allow to make.
    request_rate_per_sec = 0

    # Taken right from google.auth.helpers:
    clock_skew = timedelta(seconds=10)

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 2000

    # The 2000 characters above defined by the body_maxlen include that of the
    # title.  Setting this to True ensures overflow options behave properly
    overflow_amalgamate_title = True

    # Discord has a limit of the number of fields you can include in an
    # embeds message. This value allows the discord message to safely
    # break into multiple messages to handle these cases.
    discord_max_fields = 10

    # Define object templates
    templates = (
        "{schema}://{webhook_id}/{webhook_token}",
        "{schema}://{botname}@{webhook_id}/{webhook_token}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "botname": {
                "name": _("Bot Name"),
                "type": "string",
                "map_to": "user",
            },
            "webhook_id": {
                "name": _("Webhook ID"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "webhook_token": {
                "name": _("Webhook Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "tts": {
                "name": _("Text To Speech"),
                "type": "bool",
                "default": False,
            },
            "avatar": {
                "name": _("Avatar Image"),
                "type": "bool",
                "default": True,
            },
            "avatar_url": {
                "name": _("Avatar URL"),
                "type": "string",
            },
            "href": {
                "name": _("URL"),
                "type": "string",
            },
            "url": {
                "alias_of": "href",
            },
            # Send a message to the specified thread within a webhook's
            # channel. The thread will automatically be unarchived.
            "thread": {
                "name": _("Thread ID"),
                "type": "string",
            },
            "footer": {
                "name": _("Display Footer"),
                "type": "bool",
                "default": False,
            },
            "footer_logo": {
                "name": _("Footer Logo"),
                "type": "bool",
                "default": True,
            },
            "fields": {
                "name": _("Use Fields"),
                "type": "bool",
                "default": True,
            },
            "flags": {
                "name": _("Discord Flags"),
                "type": "int",
                "min": 0,
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": False,
                "map_to": "include_image",
            },
            "ping": {
                "name": _("Ping Users/Roles"),
                "type": "list:string",
            },
        },
    )

    def __init__(
        self,
        webhook_id,
        webhook_token,
        tts=False,
        avatar=True,
        footer=False,
        footer_logo=True,
        include_image=False,
        fields=True,
        avatar_url=None,
        href=None,
        thread=None,
        flags=None,
        ping=None,
        **kwargs,
    ):
        """Initialize Discord Object."""
        super().__init__(**kwargs)

        # Webhook ID (associated with project)
        self.webhook_id = validate_regex(webhook_id)
        if not self.webhook_id:
            msg = (
                f"An invalid Discord Webhook ID ({webhook_id}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Webhook Token (associated with project)
        self.webhook_token = validate_regex(webhook_token)
        if not self.webhook_token:
            msg = (
                "An invalid Discord Webhook Token "
                f"({webhook_token}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Text To Speech
        self.tts = tts

        # Over-ride Avatar Icon
        self.avatar = avatar

        # Place a footer
        self.footer = footer

        # include a footer_logo in footer
        self.footer_logo = footer_logo

        # Place a thumbnail image inline with the message body
        self.include_image = include_image

        # Use Fields
        self.fields = fields

        # Specified Thread ID
        self.thread_id = thread

        # Avatar URL
        # This allows a user to provide an over-ride to the otherwise
        # dynamically generated avatar url images
        self.avatar_url = avatar_url

        # A URL to have the title link to
        self.href = href

        # A URL to have the title link to
        if flags:
            try:
                self.flags = int(flags)
                if self.flags < NotifyDiscord.template_args["flags"]["min"]:
                    raise ValueError()

            except (TypeError, ValueError):
                msg = "An invalid Discord flags setting " \
                      "({}) was specified.".format(flags)
                self.logger.warning(msg)
                raise TypeError(msg) from None
        else:
            self.flags = None

        # For Tracking Purposes
        self.ratelimit_reset = datetime.now(timezone.utc).replace(tzinfo=None)

        # Default to 1.0
        self.ratelimit_remaining = 1.0

        # Ping targets (raw tokens from URL, already split by parse_list)
        self.ping = parse_list(ping)

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Discord Notification."""

        payload = {
            # Text-To-Speech
            "tts": self.tts,
            # If Text-To-Speech is set to True, then we do not want to wait
            # for the whole message before continuing. Otherwise, we wait
            "wait": self.tts is False,
        }

        if self.flags:
            # Set our flag if defined:
            payload["flags"] = self.flags

        # Acquire image_url
        image_url = self.image_url(notify_type)

        if self.avatar and (image_url or self.avatar_url):
            payload["avatar_url"] = (
                self.avatar_url if self.avatar_url else image_url
            )

        if self.user:
            # Optionally override the default username of the webhook
            payload["username"] = self.user

        # Associate our thread_id with our message
        params = {"thread_id": self.thread_id} if self.thread_id else None

        # Apply any pingable content to the payload
        payload.update(self.ping_payload(body, " ".join(self.ping)))

        if body:
            # our fields variable
            fields = []

            if self.notify_format == NotifyFormat.MARKDOWN:
                # Use embeds for payload
                payload["embeds"] = [{
                    "author": {
                        "name": self.app_id,
                        "url": self.app_url,
                    },
                    "title": title,
                    "description": body,
                    # Our color associated with our notification
                    "color": self.color(notify_type, int),
                }]

                if self.href:
                    payload["embeds"][0]["url"] = self.href

                if self.footer:
                    # Acquire logo URL
                    logo_url = self.image_url(notify_type, logo=True)

                    # Set Footer text to our app description
                    payload["embeds"][0]["footer"] = {
                        "text": self.app_desc,
                    }

                    if self.footer_logo and logo_url:
                        payload["embeds"][0]["footer"]["icon_url"] = logo_url

                if self.include_image and image_url:
                    payload["embeds"][0]["thumbnail"] = {
                        "url": image_url,
                        "height": 256,
                        "width": 256,
                    }

                if self.fields:
                    # Break titles out so that we can sort them in embeds
                    description, fields = self.extract_markdown_sections(body)

                    # Swap first entry for description
                    payload["embeds"][0]["description"] = description
                    if fields:
                        # Apply our additional parsing for a better
                        # presentation
                        payload["embeds"][0]["fields"] = fields[
                            : self.discord_max_fields
                        ]

                        # Remove entry from head of fields
                        fields = fields[self.discord_max_fields:]

            else:
                # not markdown
                payload["content"] = (
                    body if not title else f"{title}\r\n{body}"
                ) + payload.get("content", "")

            if not self._send(payload, params=params):
                # We failed to post our message
                return False

            # Process any remaining fields IF set
            if fields:
                payload["embeds"][0]["description"] = ""
                for i in range(0, len(fields), self.discord_max_fields):
                    payload["embeds"][0]["fields"] = fields[
                        i : i + self.discord_max_fields
                    ]
                    if not self._send(payload):
                        # We failed to post our message
                        return False

        if attach and self.attachment_support:
            # Update our payload; the idea is to preserve it's other detected
            # and assigned values for re-use here too
            payload.update({
                # Text-To-Speech
                "tts": False,
                # Wait until the upload has posted itself before continuing
                "wait": True,
            })

            #
            # Remove our text/title based content for attachment use
            #
            payload.pop("embeds", None)

            payload.pop("content", None)

            payload.pop("allow_mentions", None)

            #
            # Send our attachments
            #
            for attachment in attach:
                self.logger.info(
                    f"Posting Discord Attachment {attachment.name}"
                )
                if not self._send(payload, params=params, attach=attachment):
                    # We failed to post our message
                    return False

        # Otherwise return
        return True

    def _send(self, payload, attach=None, params=None, rate_limit=1, **kwargs):
        """Wrapper to the requests (post) object."""

        # Our headers
        headers = {
            "User-Agent": self.app_id,
        }

        # Construct Notify URL
        notify_url = (
            f"{self.notify_url}/{self.webhook_id}/{self.webhook_token}"
        )

        self.logger.debug(
            "Discord POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Discord Payload: {payload!s}")

        # By default set wait to None
        wait = None

        if self.ratelimit_remaining <= 0.0:
            # Determine how long we should wait for or if we should wait at
            # all. This isn't fool-proof because we can't be sure the client
            # time (calling this script) is completely synced up with the
            # Discord server.  One would hope we're on NTP and our clocks are
            # the same allowing this to role smoothly:

            now = datetime.now(timezone.utc).replace(tzinfo=None)
            if now < self.ratelimit_reset:
                # We need to throttle for the difference in seconds
                wait = abs(
                    (
                        self.ratelimit_reset - now + self.clock_skew
                    ).total_seconds()
                )

        # Always call throttle before any remote server i/o is made;
        self.throttle(wait=wait)

        # Perform some simple error checking
        if isinstance(attach, AttachBase):
            if not attach:
                # We could not access the attachment
                self.logger.error(
                    f"Could not access attachment {attach.url(privacy=True)}."
                )
                return False

            self.logger.debug(
                f"Posting Discord attachment {attach.url(privacy=True)}"
            )

        # Our attachment path (if specified)
        files = None
        try:

            # Open our attachment path if required:
            if attach:
                files = {
                    "file": (
                        attach.name,
                        # file handle is safely closed in `finally`; inline
                        # open is intentional
                        open(attach.path, "rb"),  # noqa: SIM115
                    )}

            else:
                headers["Content-Type"] = "application/json; charset=utf-8"

            r = requests.post(
                notify_url,
                params=params,
                data=payload if files else dumps(payload),
                headers=headers,
                files=files,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            # Handle rate limiting (if specified)
            try:
                # Store our rate limiting (if provided)
                self.ratelimit_remaining = float(
                    r.headers.get("X-RateLimit-Remaining")
                )
                self.ratelimit_reset = datetime.fromtimestamp(
                    int(r.headers.get("X-RateLimit-Reset")), timezone.utc
                ).replace(tzinfo=None)

            except (TypeError, ValueError):
                # This is returned if we could not retrieve this
                # information gracefully accept this state and move on
                pass

            if r.status_code not in (
                requests.codes.ok,
                requests.codes.no_content,
            ):

                # We had a problem
                status_str = NotifyBase.http_response_code_lookup(
                    r.status_code
                )

                if (
                    r.status_code == requests.codes.too_many_requests
                    and rate_limit > 0
                ):

                    # handle rate limiting
                    self.logger.warning(
                        "Discord rate limiting in effect; "
                        "blocking for %.2f second(s)",
                        self.ratelimit_remaining,
                    )

                    # Try one more time before failing
                    return self._send(
                        payload=payload,
                        attach=attach,
                        params=params,
                        rate_limit=rate_limit - 1,
                        **kwargs,
                    )

                self.logger.warning(
                    "Failed to send {}to Discord notification: "
                    "{}{}error={}.".format(
                        attach.name if attach else "",
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info(
                    "Sent Discord {}.".format(
                        "attachment" if attach else "notification"
                    )
                )

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred posting {}to Discord.".format(
                    attach.name if attach else ""
                )
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        except OSError as e:
            self.logger.warning(
                "An I/O error occurred while reading {}.".format(
                    attach.name if attach else "attachment"
                )
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            return False

        finally:
            # Close our file (if it's open) stored in the second element
            # of our files tuple (index 1)
            if files:
                files["file"][1].close()

        return True

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "tts": "yes" if self.tts else "no",
            "avatar": "yes" if self.avatar else "no",
            "footer": "yes" if self.footer else "no",
            "footer_logo": "yes" if self.footer_logo else "no",
            "image": "yes" if self.include_image else "no",
            "fields": "yes" if self.fields else "no",
        }

        if self.avatar_url:
            params["avatar_url"] = self.avatar_url

        if self.flags:
            params["flags"] = str(self.flags)

        if self.href:
            params["href"] = self.href

        if self.thread_id:
            params["thread"] = self.thread_id

        if self.ping:
            params["ping"] = ",".join(self.ping)

        # Ensure our botname is set
        botname = f"{self.user}@" if self.user else ""

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return (
            "{schema}://{botname}{webhook_id}/{webhook_token}/?{params}"
            .format(
                schema=self.secure_protocol,
                botname=botname,
                webhook_id=self.pprint(self.webhook_id, privacy, safe=""),
                webhook_token=self.pprint(
                    self.webhook_token, privacy, safe=""
                ),
                params=NotifyDiscord.urlencode(params),
            )
        )

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.webhook_id, self.webhook_token)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object.

        Syntax:
          discord://webhook_id/webhook_token
        """
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Store our webhook ID
        webhook_id = NotifyDiscord.unquote(results["host"])

        # Now fetch our tokens
        try:
            webhook_token = NotifyDiscord.split_path(results["fullpath"])[0]

        except IndexError:
            # Force some bad values that will get caught
            # in parsing later
            webhook_token = None

        results["webhook_id"] = webhook_id
        results["webhook_token"] = webhook_token

        # Text To Speech
        results["tts"] = parse_bool(results["qsd"].get("tts", False))

        # Use sections
        # effectively detect multiple fields and break them off
        # into sections
        results["fields"] = parse_bool(results["qsd"].get("fields", True))

        # Use Footer
        results["footer"] = parse_bool(results["qsd"].get("footer", False))

        # Use Footer Logo
        results["footer_logo"] = parse_bool(
            results["qsd"].get("footer_logo", True)
        )

        # Update Avatar Icon
        results["avatar"] = parse_bool(results["qsd"].get("avatar", True))

        # Boolean to include an image or not
        results["include_image"] = parse_bool(
            results["qsd"].get(
                "image", NotifyDiscord.template_args["image"]["default"]
            )
        )

        if "botname" in results["qsd"]:
            # Alias to User
            results["user"] = NotifyDiscord.unquote(results["qsd"]["botname"])

        if "flags" in results["qsd"]:
            # Alias to User
            results["flags"] = NotifyDiscord.unquote(results["qsd"]["flags"])

        # Extract avatar url if it was specified
        if "avatar_url" in results["qsd"]:
            results["avatar_url"] = NotifyDiscord.unquote(
                results["qsd"]["avatar_url"]
            )

        # Extract url if it was specified
        if "href" in results["qsd"]:
            results["href"] = NotifyDiscord.unquote(results["qsd"]["href"])

        elif "url" in results["qsd"]:
            results["href"] = NotifyDiscord.unquote(results["qsd"]["url"])
            # Markdown is implied
            results["format"] = NotifyFormat.MARKDOWN

        # Extract thread id if it was specified
        if "thread" in results["qsd"]:
            results["thread"] = NotifyDiscord.unquote(results["qsd"]["thread"])
            # Markdown is implied
            results["format"] = NotifyFormat.MARKDOWN

        # Extract ping targets, comma/space separated
        if "ping" in results["qsd"]:
            results["ping"] = NotifyDiscord.unquote(results["qsd"]["ping"])

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://discord.com/api/webhooks/WEBHOOK_ID/WEBHOOK_TOKEN
        Support Legacy URL as well:
            https://discordapp.com/api/webhooks/WEBHOOK_ID/WEBHOOK_TOKEN
        """

        result = re.match(
            r"^https?://discord(app)?\.com/api/webhooks/"
            r"(?P<webhook_id>[0-9]+)/"
            r"(?P<webhook_token>[A-Z0-9_-]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifyDiscord.parse_url(
                "{schema}://{webhook_id}/{webhook_token}/{params}".format(
                    schema=NotifyDiscord.secure_protocol,
                    webhook_id=result.group("webhook_id"),
                    webhook_token=result.group("webhook_token"),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )

        return None

    def ping_payload(self, *args):
        """
        Takes a body and applies the payload associated with pinging
        the users detected within
        """

        # initialize a payload object we can prepare
        payload = {}

        roles = set()
        users = set()
        parse = set()

        for arg in args:
            # parse for user id's <@123> and role IDs <@&456>
            results = USER_ROLE_DETECTION_RE.findall(arg)
            if not results:
                continue

            _content = []
            for is_role, no, value in results:
                if value:
                    parse.add(value)
                    _content.append(f"@{value}")

                elif is_role:
                    roles.add(no)
                    _content.append(f"<@&{no}>")

                else:  # is_user
                    users.add(no)
                    _content.append(f"<@{no}>")

        if not (roles or users or parse):
            # Nothing to add
            return payload

        # First time through...
        payload["allow_mentions"] = {
            "parse": list(parse),
            "users": list(users),
            "roles": list(roles),
        }

        # Add pingable elements to content field
        payload["content"] = "👉 " + " ".join(chain(
            [f"@{value}" for value in parse],
            [f"<@&{value}>" for value in roles],
            [f"<@{value}>" for value in users],
        ))

        return payload

    @staticmethod
    def extract_markdown_sections(markdown):
        """Takes a string in a markdown type format and extracts the headers
        and their corresponding sections into individual fields that get passed
        as an embed entry to Discord."""
        # Search for any header information found without it's own section
        # identifier
        match = re.match(
            r"^\s*(?P<desc>[^\s#]+.*?)(?=\s*$|[\r\n]+\s*#)",
            markdown,
            flags=re.S,
        )

        description = match.group("desc").strip() if match else ""
        if description:
            # Strip description from our string since it has been handled
            # now.
            markdown = re.sub(re.escape(description), "", markdown, count=1)

        regex = re.compile(
            r"\s*#[# \t\v]*(?P<name>[^\n]+)(\n|\s*$)"
            r"\s*((?P<value>[^#].+?)(?=\s*$|[\r\n]+\s*#))?",
            flags=re.S,
        )

        common = regex.finditer(markdown)
        fields = []
        for el in common:
            d = el.groupdict()

            fields.append({
                "name": d.get("name", "").strip("#`* \r\n\t\v"),
                "value": "```{}\n{}```".format(
                    "md" if d.get("value") else "",
                    (d.get("value").strip() + "\n" if d.get("value") else ""),
                ),
            })

        return description, fields



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import (
    is_email,
    is_phone_no,
    parse_bool,
    parse_list,
    validate_regex,
)
from .base import NotifyBase


class NotifyPopcornNotify(NotifyBase):
    """A wrapper for PopcornNotify Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "PopcornNotify"

    # The services URL
    service_url = "https://popcornnotify.com/"

    # The default protocol
    secure_protocol = "popcorn"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_popcornnotify"

    # PopcornNotify uses the http protocol
    notify_url = "https://popcornnotify.com/notify"

    # The maximum targets to include when doing batch transfers
    default_batch_size = 10

    # Define object templates
    templates = ("{schema}://{apikey}/{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "regex": (r"^[a-z0-9]+$", "i"),
                "required": True,
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(self, apikey, targets=None, batch=False, **kwargs):
        """Initialize PopcornNotify Object."""
        super().__init__(**kwargs)

        # Access Token (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"An invalid PopcornNotify API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Prepare Batch Mode Flag
        self.batch = batch

        # Parse our targets
        self.targets = []

        for target in parse_list(targets):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if result:
                # store valid phone number
                self.targets.append(result["full"])
                continue

            result = is_email(target)
            if result:
                # store valid email
                self.targets.append(result["full_email"])
                continue

            self.logger.warning(
                f"Dropped invalid target ({target}) specified.",
            )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform PopcornNotify Notification."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning(
                "There were no PopcornNotify targets to notify."
            )
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
        }

        # Prepare our payload
        payload = {
            "message": body,
            "subject": title,
        }

        auth = (self.apikey, None)

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        for index in range(0, len(self.targets), batch_size):
            # Prepare our recipients
            payload["recipients"] = ",".join(
                self.targets[index : index + batch_size]
            )

            self.logger.debug(
                "PopcornNotify POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"PopcornNotify Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    auth=auth,
                    data=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyPopcornNotify.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send {} PopcornNotify notification{}: "
                        "{}{}error={}.".format(
                            len(self.targets[index : index + batch_size]),
                            (
                                f" to {self.targets[index]}"
                                if batch_size == 1
                                else "(s)"
                            ),
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        "Sent {} PopcornNotify notification{}.".format(
                            len(self.targets[index : index + batch_size]),
                            (
                                f" to {self.targets[index]}"
                                if batch_size == 1
                                else "(s)"
                            ),
                        )
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occured sending"
                    f" {len(self.targets[index:index + batch_size])} "
                    "PopcornNotify notification(s)."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.apikey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "batch": "yes" if self.batch else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{apikey}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets="/".join(
                [NotifyPopcornNotify.quote(x, safe="") for x in self.targets]
            ),
            params=NotifyPopcornNotify.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + (
                1 if targets % batch_size else 0
            )

        return targets

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyPopcornNotify.split_path(
            results["fullpath"]
        )

        # The hostname is our authentication key
        results["apikey"] = NotifyPopcornNotify.unquote(results["host"])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyPopcornNotify.parse_list(
                results["qsd"]["to"]
            )

        # Get Batch Mode Flag
        results["batch"] = parse_bool(results["qsd"].get("batch", False))

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this plugin, you need to first access https://dev.flock.com/webhooks
# Specifically https://dev.flock.com/webhooks/incoming
#
# To create a new incoming webhook for your account. You'll need to
# follow the wizard to pre-determine the channel(s) you want your
# message to broadcast to. When you've completed this, you will
# recieve a URL that looks something like this:
# https://api.flock.com/hooks/sendMessage/134b8gh0-eba0-4fa9-ab9c-257ced0e8221
#                                                             ^
#                                                             |
#  This is important <----------------------------------------^
#
#  It becomes your 'token' that you will pass into this class
#
from json import dumps
import re

import requests

from ..common import NotifyFormat, NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool, parse_list, validate_regex
from .base import NotifyBase

# Extend HTTP Error Messages
FLOCK_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Token.",
}

# Used to detect a channel/user
IS_CHANNEL_RE = re.compile(r"^(#|g:)(?P<id>[A-Z0-9_]+)$", re.I)
IS_USER_RE = re.compile(r"^(@|u:)?(?P<id>[A-Z0-9_]+)$", re.I)


class NotifyFlock(NotifyBase):
    """A wrapper for Flock Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Flock"

    # The services URL
    service_url = "https://flock.com/"

    # The default secure protocol
    secure_protocol = "flock"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_flock"

    # Flock uses the http protocol with JSON requests
    notify_url = "https://api.flock.com/hooks/sendMessage"

    # API Wrapper
    notify_api = "https://api.flock.co/v1/chat.sendMessage"

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_72

    # Define object templates
    templates = (
        "{schema}://{token}",
        "{schema}://{botname}@{token}",
        "{schema}://{botname}@{token}/{targets}",
        "{schema}://{token}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Access Key"),
                "type": "string",
                "regex": (r"^[a-z0-9-]+$", "i"),
                "private": True,
                "required": True,
            },
            "botname": {
                "name": _("Bot Name"),
                "type": "string",
                "map_to": "user",
            },
            "to_user": {
                "name": _("To User ID"),
                "type": "string",
                "prefix": "@",
                "regex": (r"^[A-Z0-9_]+$", "i"),
                "map_to": "targets",
            },
            "to_channel": {
                "name": _("To Channel ID"),
                "type": "string",
                "prefix": "#",
                "regex": (r"^[A-Z0-9_]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(self, token, targets=None, include_image=True, **kwargs):
        """Initialize Flock Object."""
        super().__init__(**kwargs)

        # Build ourselves a target list
        self.targets = []

        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = f"An invalid Flock Access Key ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Track whether or not we want to send an image with our notification
        # or not.
        self.include_image = include_image

        # Track any issues
        has_error = False

        # Tidy our targets
        targets = parse_list(targets)

        for target in targets:
            result = IS_USER_RE.match(target)
            if result:
                self.targets.append("u:" + result.group("id"))
                continue

            result = IS_CHANNEL_RE.match(target)
            if result:
                self.targets.append("g:" + result.group("id"))
                continue

            has_error = True
            self.logger.warning(
                f"Ignoring invalid target ({target}) specified."
            )

        if has_error and not self.targets:
            # We have a bot token and no target(s) to message
            msg = "No Flock targets to notify."
            self.logger.warning(msg)
            raise TypeError(msg)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Flock Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # error tracking (used for function return)
        has_error = False

        if self.notify_format == NotifyFormat.HTML:
            body = f"<flockml>{body}</flockml>"

        else:
            title = NotifyFlock.escape_html(title, whitespace=False)
            body = NotifyFlock.escape_html(body, whitespace=False)

            body = "<flockml>{}{}</flockml>".format(
                "" if not title else f"<b>{title}</b><br/>", body
            )

        payload = {
            "token": self.token,
            "flockml": body,
            "sendAs": {
                "name": self.app_id if not self.user else self.user,
                # A Profile Image is only configured if we're configured to
                # allow it
                "profileImage": (
                    None
                    if not self.include_image
                    else self.image_url(notify_type)
                ),
            },
        }

        if len(self.targets):
            # Create a copy of our targets
            targets = list(self.targets)

            while len(targets) > 0:
                # Get our first item
                target = targets.pop(0)

                # Copy and update our payload
                _payload = payload.copy()
                _payload["to"] = target

                if not self._post(self.notify_api, headers, _payload):
                    has_error = True

        else:
            # Webhook
            url = f"{self.notify_url}/{self.token}"
            if not self._post(url, headers, payload):
                has_error = True

        return not has_error

    def _post(self, url, headers, payload):
        """A wrapper to the requests object."""

        # error tracking (used for function return)
        has_error = False

        self.logger.debug(
            f"Flock POST URL: {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Flock Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyFlock.http_response_code_lookup(
                    r.status_code, FLOCK_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send Flock notification : {}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Mark our failure
                has_error = True

            else:
                self.logger.info("Sent Flock notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Flock notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Mark our failure
            has_error = True

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{token}/{targets}?{params}".format(
            schema=self.secure_protocol,
            token=self.pprint(self.token, privacy, safe=""),
            targets="/".join(
                [NotifyFlock.quote(target, safe="") for target in self.targets]
            ),
            params=NotifyFlock.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyFlock.split_path(results["fullpath"])

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyFlock.parse_list(results["qsd"]["to"])

        # The first token is stored in the hostname
        results["token"] = NotifyFlock.unquote(results["host"])

        # Include images with our message
        results["include_image"] = parse_bool(
            results["qsd"].get("image", True)
        )

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://api.flock.com/hooks/sendMessage/TOKEN
        """

        result = re.match(
            r"^https?://api\.flock\.com/hooks/sendMessage/"
            r"(?P<token>[a-z0-9-]{24})/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifyFlock.parse_url(
                "{schema}://{token}/{params}".format(
                    schema=NotifyFlock.secure_protocol,
                    token=result.group("token"),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )

        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from itertools import chain

# To use this service you will need a Clickatell account to which you can get
# your API_TOKEN at:
#     https://www.clickatell.com/
import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_phone_no, parse_phone_no, validate_regex
from .base import NotifyBase


class NotifyClickatell(NotifyBase):
    """A wrapper for Clickatell Notifications."""

    # The default descriptive name associated with the Notification
    service_name = _("Clickatell")

    # The services URL
    service_url = "https://www.clickatell.com/"

    # All notification requests are secure
    secure_protocol = "clickatell"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_clickatell"

    # Clickatell API Endpoint
    notify_url = "https://platform.clickatell.com/messages/http/send"

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    templates = (
        "{schema}://{apikey}/{targets}",
        "{schema}://{source}@{apikey}/{targets}",
    )

    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "source": {
                "name": _("From Phone No"),
                "type": "string",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "apikey": {"alias_of": "apikey"},
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "source",
            },
        },
    )

    def __init__(self, apikey, source=None, targets=None, **kwargs):
        """Initialize Clickatell Object."""

        super().__init__(**kwargs)

        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid Clickatell API Token ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.source = None
        if source:
            result = is_phone_no(source)
            if not result:
                msg = (
                    "The Account (From) Phone # specified "
                    f"({source}) is invalid."
                )
                self.logger.warning(msg)

                raise TypeError(msg)

            # Tidy source
            self.source = result["full"]

        # Used for URL generation afterwards only
        self._invalid_targets = []

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets, prefix=True):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                self._invalid_targets.append(target)
                continue

            # store valid phone number
            self.targets.append(result["full"])

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.apikey, self.source)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{source}{apikey}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            source=f"{self.source}@" if self.source else "",
            apikey=self.pprint(self.apikey, privacy, safe="="),
            targets="/".join([
                NotifyClickatell.quote(t, safe="")
                for t in chain(self.targets, self._invalid_targets)
            ]),
            params=self.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification.

        Always return 1 at least
        """
        return len(self.targets) if self.targets else 1

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Clickatell Notification."""

        if not self.targets:
            # There were no targets to notify
            self.logger.warning("There were no Clickatell targets to notify")
            return False

        headers = {
            "User-Agent": self.app_id,
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        params_base = {
            "apiKey": self.apikey,
            "from": self.source,
            "content": body,
        }

        # error tracking (used for function return)
        has_error = False

        for target in self.targets:
            params = params_base.copy()
            params["to"] = target

            # Some Debug Logging
            self.logger.debug(
                "Clickatell GET URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"Clickatell Payload: {params}")

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.get(
                    self.notify_url,
                    params=params,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if (
                    r.status_code != requests.codes.ok
                    and r.status_code != requests.codes.accepted
                ):
                    # We had a problem
                    status_str = self.http_response_code_lookup(r.status_code)

                    self.logger.warning(
                        "Failed to send Clickatell notification: "
                        "{}{}error={}.".format(
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")
                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        "Sent Clickatell notification to %s", target
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Clickatell: to %s ",
                    target,
                )
                self.logger.debug(f"Socket Exception: {e!s}")
                # Mark our failure
                has_error = True
                continue

        return not has_error

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't parse the URL
            return results

        results["targets"] = NotifyClickatell.split_path(results["fullpath"])
        results["apikey"] = NotifyClickatell.unquote(results["host"])

        if results["user"]:
            results["source"] = NotifyClickatell.unquote(results["user"])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyClickatell.parse_phone_no(
                results["qsd"]["to"]
            )

        # Support the 'from'  and 'source' variable so that we can support
        # targets this way too.
        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyClickatell.unquote(
                results["qsd"]["from"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""
46elks SMS Notification Service.

Minimal URL formats (source ends up being target):
  - 46elks://user:pass@/+15551234567
  - 46elks://user:pass@/+15551234567/+46701234567
  - 46elks://user:pass@/+15551234567?from=Acme
"""

from __future__ import annotations

from collections.abc import Iterable
import re
from typing import Any, Optional

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import (
    is_phone_no,
    parse_phone_no,
)
from .base import NotifyBase


class Notify46Elks(NotifyBase):
    """A wrapper for 46elks Notifications."""

    # The default descriptive name associated with the Notification
    service_name = _("46elks")

    # The services URL
    service_url = "https://46elks.com"

    # The default secure protocol
    secure_protocol = ("46elks", "elks")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_46elks"

    # 46elksAPI Request URLs
    notify_url = "https://api.46elks.com/a1/sms"

    # The maximum allowable characters allowed in the title per message
    title_maxlen = 0

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 160

    # Define object templates
    templates = (
        "{schema}://{user}:{password}@/{from_phone}",
        "{schema}://{user}:{password}@/{from_phone}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("API Username"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("API Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "from_phone": {
                "name": _("From Phone No"),
                "type": "string",
                "required": True,
                "map_to": "source",
            },
            "target_phone": {
                "name": _("Target Phone"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "from_phone",
            },
        },
    )

    def __init__(
        self,
        targets: Optional[Iterable[str]] = None,
        source: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialise 46elks notifier.

        :param targets: Iterable of phone numbers. E.164 is recommended.
        :param source: Optional source ID or E.164 number.
        """
        super().__init__(**kwargs)

        # Prepare our source
        self.source: Optional[str] = (source or "").strip() or None

        if not self.password:
            msg = "No 46elks password was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        elif not self.user:
            msg = "No 46elks user was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Parse our targets
        self.targets = []

        if not targets and is_phone_no(self.source):
            targets = [self.source]

        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            # Carry forward '+' if defined, otherwise do not...
            self.targets.append(
                ("+" + result["full"])
                if target.lstrip()[0] == "+"
                else result["full"]
            )

    def send(
        self,
        body: str,
        title: str = "",
        notify_type: NotifyType = NotifyType.INFO,
        **kwargs: Any,
    ) -> bool:
        """Perform 46elks Notification."""

        if not self.targets:
            # There is no one to email; we're done
            self.logger.warning(
                "There are no 46elks recipients to notify"
            )
            return False

        headers = {
            "User-Agent": self.app_id,
        }

        # error tracking (used for function return)
        has_error = False

        targets = list(self.targets)
        while targets:
            target = targets.pop(0)

            # Prepare our payload
            payload = {
                "to": target,
                "from": self.source,
                "message": body,
            }

            self.logger.debug(
                "46elks POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"46elks Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=payload,
                    headers=headers,
                    auth=(self.user, self.password),
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = (
                        Notify46Elks.http_response_code_lookup(
                            r.status_code
                        )
                    )

                    self.logger.warning(
                        "Failed to send 46elks notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        f"Sent 46elks notification to {target}."
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending 46elks"
                    f" notification to {target}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another similar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol[0], self.user, self.password, self.source)

    def url(self, privacy: bool = False, *args: Any, **kwargs: Any) -> str:
        """Returns the URL built dynamically based on specified arguments."""

        # Initialize our parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        # Apprise URL can be condensed and target can be eliminated if its
        # our source phone no
        targets = (
            [] if len(self.targets) == 1 and
            self.source in self.targets else self.targets)

        return "{schema}://{user}:{pw}@{source}/{targets}?{params}".format(
            schema=self.secure_protocol[0],
            user=self.quote(self.user, safe=""),
            source=self.source if self.source else "",
            pw=self.pprint(
                self.password, privacy, mode=PrivacyMode.Secret, safe=""),
            targets="/".join(
                [Notify46Elks.quote(x, safe="+") for x in targets]
            ),
            params=Notify46Elks.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_native_url(url):
        """
        Support https://user:pw@api.46elks.com/a1/sms?to=+15551234567&from=Acme
        """

        result = re.match(
            r"^https?://(?P<credentials>[^@]+)@"
            r"api\.46elks\.com/a1/sms/?"
            r"(?P<params>\?.+)$",
            url,
            re.I,
        )

        if result:
            return Notify46Elks.parse_url(
                "{schema}://{credentials}@/{params}".format(
                    schema=Notify46Elks.secure_protocol[0],
                    credentials=result.group("credentials"),
                    params=result.group("params"),
                )
            )

        return None

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Prepare our targets
        results["targets"] = []

        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = Notify46Elks.unquote(
                results["qsd"]["from"]
            )

        elif results["host"]:
            results["source"] = Notify46Elks.unquote(results["host"])

        # Store our remaining targets found on path
        results["targets"].extend(
            Notify46Elks.split_path(results["fullpath"])
        )

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += Notify46Elks.parse_phone_no(
                results["qsd"]["to"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import binascii
import json
import os
import struct
from typing import Optional, Union

from ..apprise_attachment import AppriseAttachment
from ..asset import AppriseAsset
from ..exception import ApprisePluginException
from ..logger import logger
from ..utils.base64 import base64_urldecode, base64_urlencode

try:
    from cryptography.exceptions import InvalidTag
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes, serialization
    from cryptography.hazmat.primitives.asymmetric import ec
    from cryptography.hazmat.primitives.asymmetric.utils import (
        decode_dss_signature,
    )
    from cryptography.hazmat.primitives.ciphers import (
        Cipher,
        algorithms,
        modes,
    )
    from cryptography.hazmat.primitives.ciphers.aead import AESGCM
    from cryptography.hazmat.primitives.kdf.hkdf import HKDF
    from cryptography.hazmat.primitives.serialization import (
        Encoding,
        NoEncryption,
        PrivateFormat,
        PublicFormat,
    )

    # PEM Support enabled
    PEM_SUPPORT = True

except ImportError:
    # PEM Support disabled
    PEM_SUPPORT = False


class ApprisePEMException(ApprisePluginException):
    """Thrown when there is an error with the PEM Controller."""

    def __init__(self, message, error_code=612):
        super().__init__(message, error_code=error_code)


class ApprisePEMController:
    """PEM Controller Tool for the Apprise Library."""

    # There is no reason a PEM Public Key should exceed 8K in size
    # If it is more than this, then it is not accepted
    max_pem_public_key_size = 8000

    # There is no reason a PEM Private Key should exceed 8K in size
    # If it is more than this, then it is not accepted
    max_pem_private_key_size = 8000

    # Maximum Vapid Message Size
    max_webpush_record_size = 4096

    def __init__(
        self,
        path: str,
        pub_keyfile: Optional[str] = None,
        prv_keyfile: Optional[str] = None,
        name: Optional[str] = None,
        asset: Optional[AppriseAsset] = None,
        **kwargs,
    ) -> None:
        """Path should be the directory keys can be written and read from such
        as <notifyobject>.store.path.

        Optionally additionally specify a keyfile to explicitly open
        """

        # Directory we can work with
        self.path = path

        # Prepare our Key Placeholders
        self.__private_key = None
        self.__public_key = None

        # Our name (id)
        self.name = (
            name.strip(" \t/-+!$@#*").lower() if isinstance(name, str) else ""
        )

        # Prepare our Asset Object
        self.asset = (
            asset if isinstance(asset, AppriseAsset) else AppriseAsset()
        )

        # Our temporary reference points
        self._prv_keyfile = AppriseAttachment(asset=self.asset)
        self._pub_keyfile = AppriseAttachment(asset=self.asset)

        if prv_keyfile:
            self.load_private_key(prv_keyfile)

        elif pub_keyfile:
            self.load_public_key(pub_keyfile)

        else:
            self._pub_keyfile = None

    def load_private_key(
        self, path: Optional[str] = None, *names: str
    ) -> bool:
        """Load Private key and from that we can prepare our public key."""

        if path is None:
            # Auto-load our content
            return bool(self.private_keyfile(*names))

        # Create ourselves an Attachment to work with; this grants us the
        # ability to pull this key from a remote site or anything else
        # supported by the Attachment object
        self._prv_keyfile = AppriseAttachment(asset=self.asset)

        # Add our definition to our pem_key reference
        self._prv_keyfile.add(path)

        # Enforce maximum file size
        self._prv_keyfile[0].max_file_size = self.max_pem_private_key_size

        #
        # Reset Public key
        #
        self._pub_keyfile = AppriseAttachment(asset=self.asset)

        #
        # Reset our internal keys
        #
        self.__private_key = None
        self.__public_key = None

        if not self._prv_keyfile.sync():
            # Early exit
            logger.error(f"Could not access PEM Private Key {path}.")
            return False

        try:
            with open(self._prv_keyfile[0].path, "rb") as f:
                self.__private_key = serialization.load_pem_private_key(
                    f.read(),
                    password=None,  # or provide the password if encrypted
                    backend=default_backend(),
                )

        except (ValueError, TypeError):
            logger.debug(
                "PEM Private Key file specified is not supported (%s)",
                type(path),
            )
            return False

        except FileNotFoundError:
            logger.debug("PEM Private Key file not found: %s", path)
            return False

        except OSError as e:
            logger.warning("Error accessing PEM Private Key file %s", path)
            logger.debug(f"I/O Exception: {e}")
            return False

        #
        # Generate our public key
        #
        self.__public_key = self.__private_key.public_key()

        # Load our private key
        return bool(self.__private_key)

    def load_public_key(self, path: Optional[str] = None, *names: str) -> bool:
        """Load Public key only.

        Note: with just a public key you can only decrypt, encryption is not
              possible.
        """

        if path is None:
            # Auto-load our content
            return bool(self.public_keyfile(*names))

        # Create ourselves an Attachment to work with; this grants us the
        # ability to pull this key from a remote site or anything else
        # supported by the Attachment object
        self._pub_keyfile = AppriseAttachment(asset=self.asset)

        # Add our definition to our pem_key reference
        self._pub_keyfile.add(path)

        # Enforce maximum file size
        self._pub_keyfile[0].max_file_size = self.max_pem_public_key_size

        #
        # Reset Private key
        #
        self._prv_keyfile = AppriseAttachment(asset=self.asset)

        #
        # Reset our internal keys
        #
        self.__private_key = None
        self.__public_key = None

        if not self._pub_keyfile.sync():
            # Early exit
            logger.error(f"Could not access PEM Public Key {path}.")
            return False

        try:
            with open(path, "rb") as key_file:
                self.__public_key = serialization.load_pem_public_key(
                    key_file.read(), backend=default_backend()
                )

        except (ValueError, TypeError):
            logger.debug(
                "PEM Public Key file specified is not supported (%s)",
                type(path),
            )
            return False

        except FileNotFoundError:
            # Generate keys
            logger.debug("PEM Public Key file not found: %s", path)
            return False

        except OSError as e:
            logger.warning("Error accessing PEM Public Key file %s", path)
            logger.debug(f"I/O Exception: {e}")
            return False

        # Load our private key
        return bool(self.__public_key)

    def keygen(self, name: "Optional[str]" = None, force: bool = False):
        """Generates a set of keys based on name configured."""

        if not PEM_SUPPORT:
            msg = "PEM Support unavailable; install cryptography library"
            logger.warning(msg)
            raise ApprisePEMException(msg)

        # Detect if a key has been loaded or not
        has_key = bool(
            self.private_key(autogen=False) or self.public_key(autogen=False)
        )

        if (has_key and not (name or force)) or not self.path:
            logger.trace(
                "PEM keygen disabled, reason=%s",
                "keyfile-defined" if not has_key else "no-write-path",
            )
            return False

        # Create a new private/public key pair
        self.__private_key = ec.generate_private_key(
            ec.SECP256R1(), default_backend()
        )
        self.__public_key = self.__private_key.public_key()

        #
        # Prepare our PEM formatted output files
        #
        private_key = self.__private_key.private_bytes(
            Encoding.PEM,
            PrivateFormat.PKCS8,
            encryption_algorithm=NoEncryption(),
        )

        public_key = self.__public_key.public_bytes(
            encoding=Encoding.PEM,
            format=PublicFormat.SubjectPublicKeyInfo,
        )

        if not name:
            name = self.name

        file_prefix = "" if not name else f"{name}-"
        pub_path = os.path.join(self.path, f"{file_prefix}public_key.pem")
        prv_path = os.path.join(self.path, f"{file_prefix}private_key.pem")

        if not force:
            if os.path.isfile(pub_path):
                logger.debug(
                    "PEM generation skipped; Public Key already exists: %s/%s",
                    os.path.dirname(pub_path),
                    os.path.basename(pub_path),
                )
                return False

            if os.path.isfile(prv_path):
                logger.debug(
                    "PEM generation skipped; Private Key already exists: %s%s",
                    os.path.dirname(prv_path),
                    os.path.basename(prv_path),
                )
                return False

        try:
            # Write our keys to disk
            with open(pub_path, "wb") as f:
                f.write(public_key)

        except OSError as e:
            logger.warning("Error writing Public PEM file %s", pub_path)
            logger.debug(f"I/O Exception: {e}")

            # Cleanup
            try:
                os.unlink(pub_path)
                logger.trace("Removed %s", pub_path)

            except OSError:
                pass

            return False

        try:
            with open(prv_path, "wb") as f:
                f.write(private_key)

        except OSError as e:
            logger.warning("Error writing Private PEM file %s", prv_path)
            logger.debug(f"I/O Exception: {e}")

            try:
                os.unlink(pub_path)
                logger.trace("Removed %s", pub_path)

            except OSError:
                pass

            try:
                os.unlink(prv_path)
                logger.trace("Removed %s", prv_path)

            except OSError:
                pass

            return False

        # Update our local file references
        self._prv_keyfile = AppriseAttachment(asset=self.asset)
        self._prv_keyfile.add(prv_path)

        self._pub_keyfile = AppriseAttachment(asset=self.asset)
        self._pub_keyfile.add(pub_path)

        logger.info(
            "Wrote Public/Private PEM key pair for %s/%s",
            os.path.dirname(pub_path),
            os.path.basename(pub_path),
        )
        return True

    def public_keyfile(self, *names: str) -> Optional[str]:
        """Returns the first match of a useable public key based names
        provided."""

        if not PEM_SUPPORT:
            msg = "PEM Support unavailable; install cryptography library"
            logger.warning(msg)
            raise ApprisePEMException(msg)

        if self._pub_keyfile:
            # If our code reaches here, then we fetch our public key
            pem_key = self._pub_keyfile[0]
            if not pem_key:
                # We could not access the attachment
                logger.error(
                    "Could not access PEM Public Key"
                    f" {pem_key.url(privacy=True)}."
                )
                return False

            return pem_key.path

        elif not self.path:
            # No path
            return None

        fnames = [
            "public_key.pem",
            "public.pem",
            "pub.pem",
        ]

        if self.name:
            # Include our name in the list
            fnames = [self.name, *names]

        for name in names:
            fnames.insert(0, f"{name}-public_key.pem")

            _entry = name.lower()
            fnames.insert(0, f"{_entry}-public_key.pem")

        return next(
            (
                os.path.join(self.path, fname)
                for fname in fnames
                if os.path.isfile(os.path.join(self.path, fname))
            ),
            None,
        )

    def private_keyfile(self, *names: str) -> Optional[str]:
        """Returns the first match of a useable private key based names
        provided."""

        if not PEM_SUPPORT:
            msg = "PEM Support unavailable; install cryptography library"
            logger.warning(msg)
            raise ApprisePEMException(msg)

        if self._prv_keyfile:
            # If our code reaches here, then we fetch our private key
            pem_key = self._prv_keyfile[0]
            if not pem_key:
                # We could not access the attachment
                logger.error(
                    "Could not access PEM Private Key"
                    f" {pem_key.url(privacy=True)}."
                )
                return False

            return pem_key.path

        elif not self.path:
            # No path
            return None

        fnames = [
            "private_key.pem",
            "private.pem",
            "prv.pem",
        ]

        if self.name:
            # Include our name in the list
            fnames = [self.name, *names]

        for name in names:
            fnames.insert(0, f"{name}-private_key.pem")

            _entry = name.lower()
            fnames.insert(0, f"{_entry}-private_key.pem")

        return next(
            (
                os.path.join(self.path, fname)
                for fname in fnames
                if os.path.isfile(os.path.join(self.path, fname))
            ),
            None,
        )

    def public_key(
        self,
        *names: str,
        autogen: Optional[bool] = None,
        autodetect: bool = True,
    ) -> Optional["ec.EllipticCurvePublicKey"]:
        """Opens a spcified pem public file and returns the key from it which
        is used to decrypt the message."""
        if self.__public_key or not autodetect:
            return self.__public_key

        path = self.public_keyfile(*names)
        if not path:
            if (
                autogen if autogen is not None else self.asset.pem_autogen
            ) and self.keygen(*names):
                path = self.public_keyfile(*names)
                if path:
                    # We should get a hit now
                    return self.public_key(autogen=False)

            logger.warning("No PEM Public Key could be loaded")
            return None

        return (
            self.__public_key
            if (
                self.load_public_key(path)
                or
                # Try to see if we can load a private key (which we ca
                # generate a public from)
                self.private_key(*names, autogen=autogen)
            )
            else None
        )

    def private_key(
        self,
        *names: str,
        autogen: Optional[bool] = None,
        autodetect: bool = True,
    ) -> Optional["ec.EllipticCurvePrivateKey"]:
        """Opens a spcified pem private file and returns the key from it which
        is used to encrypt the message."""
        if self.__private_key or not autodetect:
            return self.__private_key

        path = self.private_keyfile(*names)
        if not path:
            if (
                autogen if autogen is not None else self.asset.pem_autogen
            ) and self.keygen(*names):
                path = self.private_keyfile(*names)
                if path:
                    # We should get a hit now
                    return self.private_key(autogen=False)

            logger.warning("No PEM Private Key could be loaded")
            return None

        return self.__private_key if self.load_private_key(path) else None

    def encrypt_webpush(
        self,
        message: Union[str, bytes],
        public_key: "ec.EllipticCurvePublicKey",
        auth_secret: bytes,
    ) -> bytes:
        """Encrypt a WebPush message using the recipient's public key and auth
        secret.

        Accepts input message as str or bytes.
        """
        if isinstance(message, str):
            message = message.encode("utf-8")

        # 1. Generate ephemeral EC private/Public key
        ephemeral_private_key = ec.generate_private_key(
            ec.SECP256R1(), default_backend()
        )
        ephemeral_public_key = ephemeral_private_key.public_key().public_bytes(
            encoding=Encoding.X962, format=PublicFormat.UncompressedPoint
        )

        # 2. Random salt
        salt = os.urandom(16)

        # 3. Generate shared secret via ECDH
        shared_secret = ephemeral_private_key.exchange(ec.ECDH(), public_key)

        # 4. Derive PRK using HKDF (first phase)
        recipient_public_key_bytes = public_key.public_bytes(
            encoding=Encoding.X962,
            format=PublicFormat.UncompressedPoint,
        )

        # 5. Derive Encryption key
        hkdf_secret = HKDF(
            algorithm=hashes.SHA256(),
            length=32,
            salt=auth_secret,
            info=b"WebPush: info\x00"
            + recipient_public_key_bytes
            + ephemeral_public_key,
            backend=default_backend(),
        ).derive(shared_secret)

        # 6. Derive Content Encryption Key
        hkdf_key = HKDF(
            algorithm=hashes.SHA256(),
            length=16,
            salt=salt,
            info=b"Content-Encoding: aes128gcm\x00",
            backend=default_backend(),
        ).derive(hkdf_secret)

        # 7. Derive Nonce
        hkdf_nonce = HKDF(
            algorithm=hashes.SHA256(),
            length=12,
            salt=salt,
            info=b"Content-Encoding: nonce\x00",
            backend=default_backend(),
        ).derive(hkdf_secret)

        # 8. Encrypt the message
        aesgcm = AESGCM(hkdf_key)
        # RFC8291 requires us to add '\0x02' byte to end of message
        ciphertext = aesgcm.encrypt(
            hkdf_nonce, message + b"\x02", associated_data=None
        )

        # 9. Build WebPush header + payload
        header = salt
        header += struct.pack("!L", self.max_webpush_record_size)
        header += struct.pack("!B", len(ephemeral_public_key))
        header += ephemeral_public_key
        header += ciphertext

        return header

    def encrypt(
        self,
        message: Union[str, bytes],
        public_key: "Optional[ec.EllipticCurvePublicKey]" = None,
        salt: Optional[bytes] = None,
    ) -> Optional[str]:
        """Encrypts a message using the recipient's public key (or self public
        key if none provided).

        Message can be str or bytes.
        """

        if not PEM_SUPPORT:
            msg = "PEM Support unavailable; install cryptography library"
            logger.warning(msg)
            raise ApprisePEMException(msg)

        # 1. Handle string vs bytes input
        if isinstance(message, str):
            message = message.encode("utf-8")

        # 2. Select public key
        if public_key is None:
            public_key = self.public_key()
            if public_key is None:
                logger.debug("No public key available for encryption.")
                return None

        # 3. Generate ephemeral EC private key
        ephemeral_private_key = ec.generate_private_key(
            ec.SECP256R1(), default_backend()
        )

        # 4. Derive shared secret
        shared_secret = ephemeral_private_key.exchange(ec.ECDH(), public_key)

        # 5. Derive symmetric AES key using HKDF
        derived_key = HKDF(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,  # Allow salt=None if not provided
            info=b"ecies-encryption",
            backend=default_backend(),
        ).derive(shared_secret)

        # 6. Encrypt the message using AES-GCM
        iv = os.urandom(12)  # 96-bit random IV for GCM
        encryptor = Cipher(
            algorithms.AES(derived_key),
            modes.GCM(iv),
            backend=default_backend(),
        ).encryptor()

        ciphertext = encryptor.update(message) + encryptor.finalize()
        tag = encryptor.tag

        # 7. Serialize ephemeral public key as X9.62 Uncompressed Point
        ephemeral_public_key_bytes = (
            ephemeral_private_key.public_key().public_bytes(
                encoding=serialization.Encoding.X962,
                format=serialization.PublicFormat.UncompressedPoint,
            )
        )

        # 8. Combine everything cleanly
        full_payload = {
            "ephemeral_pubkey": base64_urlencode(ephemeral_public_key_bytes),
            "iv": base64_urlencode(iv),
            "tag": base64_urlencode(tag),
            "ciphertext": base64_urlencode(ciphertext),
        }

        return base64.b64encode(
            json.dumps(full_payload).encode("utf-8")
        ).decode("utf-8")

    def decrypt(
        self,
        encrypted_payload: Union[str, bytes],
        private_key: "Optional[ec.EllipticCurvePrivateKey]" = None,
        salt: Optional[bytes] = None,
    ) -> Optional[str]:
        """Decrypts a message using the provided private key or fallback to
        self's private key.

        Payload is the base64-encoded JSON from encrypt().
        """

        if not PEM_SUPPORT:
            msg = "PEM Support unavailable; install cryptography library"
            logger.warning(msg)
            raise ApprisePEMException(msg)

        # 1. Parse input
        try:
            if isinstance(encrypted_payload, str):
                payload_bytes = base64.b64decode(
                    encrypted_payload.encode("utf-8")
                )

            else:
                payload_bytes = base64.b64decode(encrypted_payload)

        except binascii.Error:
            # Bad Padding
            logger.debug("Unparseable encrypted content provided")
            return None

        try:
            payload = json.loads(payload_bytes.decode("utf-8"))

        except UnicodeDecodeError:
            logger.debug("Unparseable encrypted content provided")
            return None

        ephemeral_pubkey_bytes = base64_urldecode(payload["ephemeral_pubkey"])
        iv = base64_urldecode(payload["iv"])
        tag = base64_urldecode(payload["tag"])
        ciphertext = base64_urldecode(payload["ciphertext"])

        # 2. Select private key
        if private_key is None:
            private_key = self.private_key()
            if private_key is None:
                logger.debug("No private key available for decryption.")
                return None

        # 3. Load ephemeral public key from sender
        ephemeral_pubkey = ec.EllipticCurvePublicKey.from_encoded_point(
            ec.SECP256R1(), ephemeral_pubkey_bytes
        )

        # 4. ECDH shared secret
        shared_secret = private_key.exchange(ec.ECDH(), ephemeral_pubkey)

        # 5. Derive symmetric AES key with HKDF
        derived_key = HKDF(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            info=b"ecies-encryption",
            backend=default_backend(),
        ).derive(shared_secret)

        # 6. Decrypt using AES-GCM
        decryptor = Cipher(
            algorithms.AES(derived_key),
            modes.GCM(iv, tag),
            backend=default_backend(),
        ).decryptor()

        try:
            plaintext = decryptor.update(ciphertext) + decryptor.finalize()

        except InvalidTag:
            logger.debug("Decryption failed - Authentication Mismatch")
            # Reason for Error:
            #   - Mismatched or missing salt
            #   - Mismatched iv, tag, or ciphertext
            #   - Incorrect or corrupted ephemeral_pubkey
            #   - Wrong or incomplete key derivation
            #   - Data being altered between encryption and decryption
            #     (truncated/corrupted)

            # Basically if we get here, we tried to decrypt encrypted content
            # using the wrong key.
            return None

        # 7. Return decoded message
        return plaintext.decode("utf-8")

    def sign(self, content: bytes) -> Optional[bytes]:
        """Sign the message using ES256 (ECDSA w/ SHA256) via private key."""

        try:
            # Sign the message using ES256 (ECDSA w/ SHA256)
            der_sig = self.private_key().sign(
                content, ec.ECDSA(hashes.SHA256())
            )

        except AttributeError:
            # NoneType; could not load key
            return None

        # Convert DER to raw R||S
        r, s = decode_dss_signature(der_sig)
        return r.to_bytes(32, byteorder="big") + s.to_bytes(
            32, byteorder="big"
        )

    @property
    def pub_keyfile(self) -> Optional[Union[str, bool]]:
        """Returns the Public Keyfile Path if set otherwise it returns None
        This property returns False if a keyfile was provided, but was
        invalid."""
        return (
            None
            if not self._pub_keyfile
            else (
                False
                if not self._pub_keyfile[0]
                else self._pub_keyfile[0].path
            )
        )

    @property
    def prv_keyfile(self) -> Optional[Union[str, bool]]:
        """Returns the Private Keyfile Path if set otherwise it returns None
        This property returns False if a keyfile was provided, but was
        invalid."""
        return (
            None
            if not self._prv_keyfile
            else (
                False
                if not self._prv_keyfile[0]
                else self._prv_keyfile[0].path
            )
        )

    @property
    def x962_str(self) -> str:
        """X962 serialization based on public key."""
        try:
            return base64_urlencode(
                self.public_key().public_bytes(
                    encoding=serialization.Encoding.X962,
                    format=serialization.PublicFormat.UncompressedPoint,
                )
            )
        except AttributeError:
            # Public Key could not be generated (public_key() returned None)
            return ""

    def __bool__(self) -> bool:
        """Returns True if at least 1 key was loaded."""
        return bool(self.private_key() or self.public_key())



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# For this plugin to work, you need to add the Maker applet to your profile
# Simply visit https://ifttt.com/search and search for 'Webhooks'
# Or if you're signed in, click here: https://ifttt.com/maker_webhooks
# and click 'Connect'
#
# You'll want to visit the settings of this Applet and pay attention to the
# URL. For example, it might look like this:
#               https://maker.ifttt.com/use/a3nHB7gA9TfBQSqJAHklod
#
# In the above example a3nHB7gA9TfBQSqJAHklod becomes your {webhook_id}
# You will need this to make this notification work correctly
#
# For each event you create you will assign it a name (this will be known as
# the {event} when building your URL.
from json import dumps
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase


class NotifyIFTTT(NotifyBase):
    """A wrapper for IFTTT Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "IFTTT"

    # The services URL
    service_url = "https://ifttt.com/"

    # The default protocol
    secure_protocol = "ifttt"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_ifttt"

    # Even though you'll add 'Ingredients' as {{ Value1 }} to your Applets,
    # you must use their lowercase value in the HTTP POST.
    ifttt_default_key_prefix = "value"

    # The default IFTTT Key to use when mapping the title text to the IFTTT
    # event. The idea here is if someone wants to over-ride the default and
    # change it to another Ingredient Name (in 2018, you were limited to have
    # value1, value2, and value3).
    ifttt_default_title_key = "value1"

    # The default IFTTT Key to use when mapping the body text to the IFTTT
    # event. The idea here is if someone wants to over-ride the default and
    # change it to another Ingredient Name (in 2018, you were limited to have
    # value1, value2, and value3).
    ifttt_default_body_key = "value2"

    # The default IFTTT Key to use when mapping the body text to the IFTTT
    # event. The idea here is if someone wants to over-ride the default and
    # change it to another Ingredient Name (in 2018, you were limited to have
    # value1, value2, and value3).
    ifttt_default_type_key = "value3"

    # IFTTT uses the http protocol with JSON requests
    notify_url = (
        "https://maker.ifttt.com/trigger/{event}/with/key/{webhook_id}"
    )

    # Define object templates
    templates = ("{schema}://{webhook_id}/{events}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "webhook_id": {
                "name": _("Webhook ID"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "events": {
                "name": _("Events"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "events",
            },
        },
    )

    # Define our token control
    template_kwargs = {
        "add_tokens": {
            "name": _("Add Tokens"),
            "prefix": "+",
        },
        "del_tokens": {
            "name": _("Remove Tokens"),
            "prefix": "-",
        },
    }

    def __init__(
        self, webhook_id, events, add_tokens=None, del_tokens=None, **kwargs
    ):
        """Initialize IFTTT Object.

        add_tokens can optionally be a dictionary of key/value pairs that you
        want to include in the IFTTT post to the server.

        del_tokens can optionally be a list/tuple/set of tokens that you want
        to eliminate from the IFTTT post.  There isn't much real functionality
        to this one unless you want to remove reference to Value1, Value2,
        and/or Value3
        """
        super().__init__(**kwargs)

        # Webhook ID (associated with project)
        self.webhook_id = validate_regex(webhook_id)
        if not self.webhook_id:
            msg = f"An invalid IFTTT Webhook ID ({webhook_id}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our Events we wish to trigger
        self.events = parse_list(events)
        if not self.events:
            msg = "You must specify at least one event you wish to trigger on."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Tokens to include in post
        self.add_tokens = {}
        if add_tokens:
            self.add_tokens.update(add_tokens)

        # Tokens to remove
        self.del_tokens = []
        if del_tokens is not None:
            if isinstance(del_tokens, (list, tuple, set)):
                self.del_tokens = del_tokens

            elif isinstance(del_tokens, dict):
                # Convert the dictionary into a list
                self.del_tokens = set(del_tokens.keys())

            else:
                msg = (
                    f"del_token must be a list; {type(del_tokens)!s} was"
                    " provided"
                )
                self.logger.warning(msg)
                raise TypeError(msg)

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform IFTTT Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # prepare JSON Object
        payload = {
            self.ifttt_default_title_key: title,
            self.ifttt_default_body_key: body,
            self.ifttt_default_type_key: notify_type,
        }

        # Add any new tokens expected (this can also potentially override
        # any entries defined above)
        payload.update(self.add_tokens)

        # Eliminate fields flagged for removal otherwise ensure all tokens are
        # lowercase since that is what the IFTTT server expects from us.
        payload = {
            x.lower(): y
            for x, y in payload.items()
            if x not in self.del_tokens
        }

        # error tracking (used for function return)
        has_error = False

        # Create a copy of our event lit
        events = list(self.events)

        while len(events):

            # Retrive an entry off of our event list
            event = events.pop(0)

            # URL to transmit content via
            url = self.notify_url.format(
                webhook_id=self.webhook_id,
                event=event,
            )

            self.logger.debug(
                "IFTTT POST URL:"
                f" {url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"IFTTT Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                self.logger.debug(
                    f"IFTTT HTTP response headers: {r.headers!r}"
                )
                self.logger.debug(f"IFTTT HTTP response body: {r.content!r}")

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyIFTTT.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send IFTTT notification to {}: "
                        "{}{}error={}.".format(
                            event,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(f"Sent IFTTT notification to {event}.")

            except requests.RequestException as e:
                self.logger.warning(
                    f"A Connection error occurred sending IFTTT:{event} "
                    + "notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.webhook_id)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        # Store any new key/value pairs added to our list
        params.update({f"+{k}": v for k, v in self.add_tokens})
        params.update({f"-{k}": "" for k in self.del_tokens})

        return "{schema}://{webhook_id}@{events}/?{params}".format(
            schema=self.secure_protocol,
            webhook_id=self.pprint(self.webhook_id, privacy, safe=""),
            events="/".join(
                [NotifyIFTTT.quote(x, safe="") for x in self.events]
            ),
            params=NotifyIFTTT.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.events)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Our API Key is the hostname if no user is specified
        results["webhook_id"] = (
            results["user"] if results["user"] else results["host"]
        )

        # Unquote our API Key
        results["webhook_id"] = NotifyIFTTT.unquote(results["webhook_id"])

        # Parse our add_token and del_token arguments (if specified)
        results["add_token"] = results["qsd+"]
        results["del_token"] = results["qsd-"]

        # Our Event
        results["events"] = []
        if results["user"]:
            # If a user was defined, then the hostname is actually a event
            # too
            results["events"].append(NotifyIFTTT.unquote(results["host"]))

        # Now fetch the remaining tokens
        results["events"].extend(NotifyIFTTT.split_path(results["fullpath"]))

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["events"] += NotifyIFTTT.parse_list(results["qsd"]["to"])

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://maker.ifttt.com/use/WEBHOOK_ID/EVENT_ID
        """

        result = re.match(
            r"^https?://maker\.ifttt\.com/use/"
            r"(?P<webhook_id>[A-Z0-9_-]+)"
            r"((?P<events>(/[A-Z0-9_-]+)+))?"
            r"/?(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifyIFTTT.parse_url(
                "{schema}://{webhook_id}{events}{params}".format(
                    schema=NotifyIFTTT.secure_protocol,
                    webhook_id=result.group("webhook_id"),
                    events=(
                        ""
                        if not result.group("events")
                        else "@{}".format(result.group("events"))
                    ),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )

        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# WeCom for PC
#   1. On WeCom for PC, find the target WeCom group for receiving alarm
#        notifications.
#   2. Right-click the WeCom group. In the window that appears, click
#        "Add Group Bot".
#   3. In the window that appears, click Create a Bot.
#   4. In the window that appears, enter a custom bot name and click Add.
#   5. You will be provided a Webhook URL that looks like:
#          https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key=abcd
#
# WeCom for Web
#   1. On WebCom for Web, open the target WeCom group for receiving alarm
#        notifications.
#   2. Click the group settings icon in the upper-right corner.
#   3. On the group settings page, choose Group Bots > Add a Bot.
#   4. On the management page for adding bots, enter a custom name for the new
#        bot.
#   5. Click Add, copy the webhook address, and configure the API callback by
#        following Step 2.

# the URL will look something like this:
#       https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key=abcd
#                                                             ^
#                                                             |
#                                                     webhook key
#
# This plugin also supports taking the URL (as identified above) directly
# as well.

from json import dumps
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase


class NotifyWeComBot(NotifyBase):
    """A wrapper for WeCom Bot Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "WeCom Bot"

    # The services URL
    service_url = "https://weixin.qq.com/"

    # The default secure protocol
    secure_protocol = "wecombot"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_wecombot"

    # Plain Text Notification URL
    notify_url = "https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key={key}"

    # Define object templates
    templates = ("{schema}://{key}",)

    # The title is not used
    title_maxlen = 0

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            # The Bot Key can be found at the end of the webhook provided
            # (?key=)
            "key": {
                "name": _("Bot Webhook Key"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[a-z0-9_-]+$", "i"),
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            # You can optionally pass IRC colors into
            "key": {
                "alias_of": "key",
            },
        },
    )

    def __init__(self, key, **kwargs):
        """Initialize WeCom Bot Object."""
        super().__init__(**kwargs)

        # Assign our bot webhook
        self.key = validate_regex(key, *self.template_tokens["key"]["regex"])
        if not self.key:
            msg = f"An invalid WeCom Bot Webhook Key ({key}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Prepare our notification URL now:
        self.api_url = self.notify_url.format(
            key=self.key,
        )
        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.key)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Prepare our parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{key}/?{params}".format(
            schema=self.secure_protocol,
            key=self.pprint(self.key, privacy, safe=""),
            params=NotifyWeComBot.urlencode(params),
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Wrapper to _send since we can alert more then one channel."""

        # prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json; charset=utf-8",
        }

        # Prepare our payload
        payload = {
            "msgtype": "text",
            "text": {
                "content": body,
            },
        }

        self.logger.debug(
            "WeCom Bot GET URL:"
            f" {self.api_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"WeCom Bot Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.api_url,
                data=dumps(payload).encode("utf-8"),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyWeComBot.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send WeCom Bot notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent WeCom Bot notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending WeCom Bot notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The first token is stored in the hostname
        results["key"] = NotifyWeComBot.unquote(results["host"])

        # The 'key' makes it easier to use yaml configuration
        if "key" in results["qsd"] and len(results["qsd"]["key"]):
            results["key"] = NotifyWeComBot.unquote(results["qsd"]["key"])

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key=BOTKEY
        """

        result = re.match(
            r"^https?://qyapi\.weixin\.qq\.com/cgi-bin/webhook/send/?\?key="
            r"(?P<key>[A-Z0-9_-]+)/?"
            r"&?(?P<params>.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifyWeComBot.parse_url(
                "{schema}://{key}{params}".format(
                    schema=NotifyWeComBot.secure_protocol,
                    key=result.group("key"),
                    params=(
                        ""
                        if not result.group("params")
                        else "?" + result.group("params")
                    ),
                )
            )

        return None
